# -*- coding: utf-8 -*-

import copy


RESERVED_KEYS = ['key', 'key_as_string', 'doc_count']


class ResultTree(object):
    def __init__(self, es_result):
        if isinstance(es_result, dict):
            self.es_result = copy.deepcopy(es_result)
        elif hasattr(es_result, '_d_'):
            self.es_result = copy.deepcopy(es_result._d_)
        else:
            raise Exception('ResultTree expects a dict or an elasticsearch_dsl Response object')

    def flatten_result(self):
        if 'aggregations' not in self.es_result:
            return []

        aggregations = self.es_result['aggregations']
        return self._extract_lines(aggregations)

    def _is_nested_node(self, node):
        # Not even a node, or a list of buckets
        if not isinstance(node, dict):
            return False

        # Standard aggregation
        if 'buckets' in node:
            return False

        # Bucket
        if 'key' in node:
            return False

        # For example the first node
        for child_node in node.values():
            if isinstance(child_node, dict)\
                and 'doc_count' in child_node\
                and not self._is_nested_node(child_node):
                return False

        # Node like {'value': 123.456}
        if all([not isinstance(child_node, dict) for child_node in node.values()]):
            return False

        return True

    def _remove_nested_aggregations(self, node):
        while True:
            new_node = self.__remove_nested_aggregations(node)
            if new_node == node:
                break
            else:
                node = new_node

        return new_node

    def __remove_nested_aggregations(self, node):
        _node = {}

        # We force an ordering to have a deterministic result
        for key, child_node in sorted(node.items(), reverse=True):
            if isinstance(child_node, dict):
                if self._is_nested_node(child_node):
                    _node.update(self._remove_nested_aggregations(child_node))
                else:
                    _node[key] = self._remove_nested_aggregations(child_node)
            elif isinstance(child_node, list):
                _node[key] = [
                    self._remove_nested_aggregations(gchild_node)
                    if isinstance(gchild_node, dict) else gchild_node
                    for gchild_node in child_node
                ]
            else:
                _node[key] = child_node

        return _node

    def _create_line(self, base_line, node):
        new_line = base_line.copy()

        for k, v in node.items():
            if k in ['key', 'key_as_string']:
                continue
            if k == 'doc_count':
                new_line[k] = v
            else:
                new_line[k] = v['value']

        return new_line

    def _is_leaf(self, node):
        # If there are still buckets, we are not on a leaf
        return 'buckets' not in node

    def _find_deeper_path(self, node):
        # The path should always end right before a buckets node, or lead to a leaf
        path = []
        current_key = None

        path.append('buckets')
        buckets = node['buckets']

        if isinstance(buckets, list):
            path.append(0)
            next_node = buckets[0]
        elif isinstance(buckets, dict):
            first_key = sorted(buckets.keys())[0]
            path.append(first_key)
            next_node = buckets[first_key]

        # We find the next key if there is one
        next_key = [k for k in next_node.keys() if k not in RESERVED_KEYS]
        if next_key:
            next_key = next_key[0]
            if 'buckets' in next_node[next_key]:
                path.append(next_key)
                current_key = next_key

        return path, current_key

    def _bootstrap_current_key(self, aggregations):
        return sorted([
            k for k in aggregations.keys()
            if k not in RESERVED_KEYS
        ])[0]

    def _extract_lines(self, aggregations):
        # Initialization
        lines = []
        depth = 0
        base_line = {}

        current_key = self._bootstrap_current_key(aggregations)
        node = aggregations[current_key]

        # Are we dealing with a metric without aggs?
        if 'buckets' not in node and 'doc_count' not in node:
            return [{
                key: aggregations[key]['value']
                for key in aggregations.keys()
            }]

        # We remove nested aggregations, I don't see the point
        # of exposing them and they are annoying to deal with
        aggregations = self._remove_nested_aggregations(aggregations)

        current_key = self._bootstrap_current_key(aggregations)
        path = [current_key]
        node = aggregations[current_key]

        while True:
            # We get the current node using the path
            node = aggregations

            for key in path:
                node = node[key]

            if self._is_leaf(node):
                # We create a new line:
                new_line = self._create_line(base_line, node)
                lines.append(new_line)

                # We delete the leaf
                parent = aggregations
                for key in path[:-1]:
                    parent = parent[key]

                del parent[path[-1]]

                # We update the path
                path.pop()
                path.pop()
                depth -= 1

                continue

            buckets = node['buckets']

            # If there are no more buckets, and we are at depth 0
            if not buckets and depth == 0:
                # If there is another level 0 aggregation, we work on it
                next_key = [
                    k for k in aggregations.keys()
                    if k not in RESERVED_KEYS and k != current_key
                ]
                if next_key:
                    aggregations.pop(current_key)
                    base_line.pop(current_key)
                    current_key = next_key[0]
                    path = [current_key]
                    continue

                # Otherwise we're done!
                break

            # If there are no more buckets but we're not at depth 0,
            # either there is another aggregation at our depth or we go higher
            if not buckets:
                base_line.pop(current_key, None)  # Buckets may have been empty from the start

                parent_bucket = aggregations
                for key in path[:-2]:
                    parent_bucket = parent_bucket[key]

                # Is there another aggregation at our level?
                next_key = [
                    k for k in parent_bucket[path[-2]].keys()
                    if k not in RESERVED_KEYS and k != current_key
                ]
                if not next_key:
                    # No, we delete whole bucket
                    del parent_bucket[path[-2]]

                    # We update the path, the depth and the current_key
                    path.pop()  # current_key
                    path.pop()  # `0` or first_key
                    path.pop()  # `buckets`
                    depth -= 1
                    current_key = path[-1]
                else:
                    # Yes, we only delete the current bucket
                    parent_bucket[path[-2]].pop(current_key)

                    # We update the path and the current_key
                    path.pop()  # current_key
                    current_key = next_key[0]
                    path.append(current_key)

                # We go again
                continue

            # We need to go one level deeper
            if isinstance(buckets, list):
                bucket = buckets[0]
                base_line.update({
                    current_key: bucket['key'],
                })
            elif isinstance(buckets, dict):
                first_key = sorted(buckets.keys())[0]
                bucket = buckets[first_key]
                base_line.update({
                    current_key: first_key,
                })

            added_path, next_key = self._find_deeper_path(node)
            path += added_path
            if next_key:
                current_key = next_key
            depth += 1

        return lines



# -*- coding: utf-8 -*-

import json
import random

from fiqs.testing.gen_data import random_timestamp, random_shop_id


def gen_traffic_data(size):
    for i in xrange(size):
        print json.dumps({
            'id': i + 1,
            'shop_id': random_shop_id(),
            'timestamp': random_timestamp(),
            'duration': random.choice([600, 900]),
            'incoming_traffic': random.randint(0, 200),
            'outgoing_traffic': random.randint(0, 200),
        })


if __name__ == '__main__':
    size = 500
    gen_traffic_data(size)



# -*- coding: utf-8 -*-

from fiqs import fields
from fiqs.models import Model


PAYMENT_TYPES = [
    'wire_transfer',
    'cash',
    'store_credit',
]


class Sale(Model):
    doc_type = 'sale'

    id = fields.IntegerField()
    shop_id = fields.IntegerField()
    client_id = fields.KeywordField()

    timestamp = fields.DateField()
    price = fields.IntegerField()
    payment_type = fields.KeywordField(choices=PAYMENT_TYPES)

    products = fields.NestedField()
    product_id = fields.KeywordField(parent='products')
    product_type = fields.KeywordField(parent='products')
    product_price = fields.IntegerField(parent='products')

    parts = fields.NestedField(parent='products')
    part_id = fields.KeywordField(parent='parts')
    warehouse_id = fields.KeywordField(parent='parts')
    part_price = fields.IntegerField(parent='parts')


class TrafficCount(Model):
    doc_type = 'traffic_count'

    id = fields.IntegerField()
    shop_id = fields.IntegerField()

    timestamp = fields.DateField()
    duration = fields.IntegerField()
    incoming_traffic = fields.IntegerField()
    outgoing_traffic = fields.IntegerField()


class SaleWithoutProducts(Model):
    doc_type = 'sale_without_products'

    id = fields.IntegerField()
    shop_id = fields.IntegerField()
    client_id = fields.KeywordField()

    timestamp = fields.DateField()
    price = fields.IntegerField()
    payment_type = fields.KeywordField(choices=PAYMENT_TYPES)


class SaleWithProducts(SaleWithoutProducts):
    doc_type = 'sale_with_products'

    products = fields.NestedField()
    product_id = fields.KeywordField(parent='products')
    product_type = fields.KeywordField(parent='products')
    product_price = fields.IntegerField(parent='products')


class SaleWithParts(SaleWithProducts):
    doc_type = 'sale_with_parts'

    parts = fields.NestedField(parent='products')
    part_id = fields.KeywordField(parent='parts')
    warehouse_id = fields.KeywordField(parent='parts')
    part_price = fields.IntegerField(parent='parts')



# -*- coding: utf-8 -*-

import six

from fiqs import flatten_result
from fiqs.tests.conftest import load_output
from fiqs.tree import ResultTree


def test_no_aggregate_no_metric():
    expected = []
    assert flatten_result(load_output('no_aggregate_no_metric')) == expected


def test_no_data():
    lines = flatten_result(load_output('no_data_nb_sales_by_day_of_week_by_shop'))
    assert lines == []


def test_nb_sales_by_shop():
    lines = flatten_result(load_output('nb_sales_by_shop'))

    assert len(lines) == 10  # One for each shop

    # Lines are sorted by doc_count
    assert lines == sorted(lines, key=(lambda l: l['doc_count']), reverse=True)

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation is present
        assert 'shop_id' in line
        assert type(line['shop_id']) == int


def test_total_sales_by_shop():
    lines = flatten_result(load_output('total_sales_by_shop'))

    assert len(lines) == 10  # One for each shop

    # Lines are sorted by doc_count
    assert lines == sorted(lines, key=(lambda l: l['doc_count']), reverse=True)

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metric are present
        assert 'shop_id' in line
        assert type(line['shop_id']) == int
        assert 'total_sales' in line
        assert type(line['total_sales']) == float


def test_total_sales_by_payment_type():
    lines = flatten_result(load_output('total_sales_by_payment_type'))

    assert len(lines) == 3  # One for each payment type

    # Lines are sorted by doc_count
    assert lines == sorted(lines, key=(lambda l: l['doc_count']), reverse=True)

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metric are present
        assert 'payment_type' in line
        assert type(line['payment_type']) == six.text_type
        assert 'total_sales' in line
        assert type(line['total_sales']) == float


def test_total_sales_by_payment_type_by_shop():
    lines = flatten_result(load_output('total_sales_by_payment_type_by_shop'))

    assert len(lines) == 3 * 10  # One for each and one for each payment type

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Both aggregations and metric are present
        assert 'payment_type' in line
        assert type(line['payment_type']) == six.text_type
        assert 'shop_id' in line
        assert type(line['shop_id']) == int
        assert 'total_sales' in line
        assert type(line['total_sales']) == float

    # Ten lines for each payment type
    for payment_type in ['cash', 'wire_transfer', 'store_credit', ]:
        payment_type_lines = [l for l in lines if l['payment_type'] == payment_type]
        assert len(payment_type_lines) == 10

        # For each payment type, lines are ordered by doc_count
        assert payment_type_lines == sorted(
            payment_type_lines, key=(lambda l: l['doc_count']), reverse=True)


def test_total_sales_day_by_day():
    lines = flatten_result(load_output('total_sales_day_by_day'))

    assert len(lines) == 31  # Number of days in the data's month

    # Lines are sorted by date
    assert lines == sorted(lines, key=(lambda l: l['timestamp']))

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # The aggregation and the metric are present
        assert 'timestamp' in line
        assert type(line['timestamp']) == int
        assert 'total_sales' in line
        assert type(line['total_sales']) == float


def test_total_sales_day_by_day_by_shop_and_by_payment():
    lines = flatten_result(load_output('total_sales_day_by_day_by_shop_and_by_payment'))

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Day by day aggregation is always present
        assert 'timestamp' in line
        assert type(line['timestamp']) == int
        # Metric is always present
        assert 'total_sales' in line
        assert type(line['total_sales']) == float
        # Either shop aggregation or payment type aggregation is present
        assert ('shop_id' in line and 'payment_type' not in line)\
            or ('payment_type' in line and 'shop_id' not in line)
        if 'shop_id' in line:
            assert type(line['shop_id']) == int
        elif 'payment_type' in line:
            assert type(line['payment_type']) == six.text_type

    # Documents are counted once in the payment aggregations, once in the shop aggregation
    sum([line['doc_count']]) == 2 * 500

    # There are more than 31 buckets within the shop buckets
    assert len([l for l in lines if 'shop_id' in l]) >= 31
    # There are more than 31 buckets within the payment buckets
    assert len([l for l in lines if 'payment_type' in l]) >= 31


def test_total_and_avg_sales_by_shop():
    lines = flatten_result(load_output('total_and_avg_sales_by_shop'))

    assert len(lines) == 10  # One for each shop

    # Lines are sorted by doc_count
    assert lines == sorted(lines, key=(lambda l: l['doc_count']), reverse=True)

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metrics are present
        assert 'shop_id' in line
        assert type(line['shop_id']) == int
        assert 'total_sales' in line
        assert type(line['total_sales']) == float
        assert 'avg_sales' in line
        assert type(line['avg_sales']) == float


def test_total_sales_by_shop_and_by_payment():
    lines = flatten_result(load_output('total_sales_by_shop_and_by_payment'))

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Metric is always present
        assert 'total_sales' in line
        assert type(line['total_sales']) == float
        # Either shop aggregation or payment type aggregation is present
        assert ('shop_id' in line and 'payment_type' not in line)\
            or ('payment_type' in line and 'shop_id' not in line)
        if 'shop_id' in line:
            assert type(line['shop_id']) == int
        elif 'payment_type' in line:
            assert type(line['payment_type']) == six.text_type

    # There are 3 payment lines, sorted by doc_count
    payment_lines = [l for l in lines if 'payment_type' in l]
    assert len(payment_lines) == 3
    assert sorted(payment_lines, key=lambda l: l['doc_count'], reverse=True) == payment_lines

    # There are 10 shop lines, sorted by doc_count
    shop_lines = [l for l in lines if 'shop_id' in l]
    assert len(shop_lines) == 10
    assert sorted(shop_lines, key=lambda l: l['doc_count'], reverse=True) == shop_lines


def test_total_sales():
    lines = flatten_result(load_output('total_sales'))

    assert len(lines) == 1

    line = lines[0]
    # Only metric is present
    assert list(line.keys()) == ['total_sales']
    assert type(line['total_sales']) == float


def test_total_sales_and_avg_sales():
    lines = flatten_result(load_output('total_sales_and_avg_sales'))

    assert len(lines) == 1

    line = lines[0]
    # Both metrics are present
    assert sorted(list(line.keys())) == sorted(['total_sales', 'avg_sales'])
    for key in ['total_sales', 'avg_sales']:
        assert type(line[key]) == float


##########
# Nested #
##########

def test_is_nested_node():
    node = {
        'products': {
            'doc_count': 100,
            'product': {
                'buckets': [
                    {
                        "doc_count": 38,
                        "key": "product_JQ0JVW20PV",
                        "parts": {
                            "doc_count": 100,
                            "part": {
                                "buckets": [
                                    {
                                        "doc_count": 10,
                                        "key": "part_10",
                                        "avg_part_price": {
                                            "value": 10.11
                                        },
                                    },
                                ],
                                "doc_count_error_upper_bound": 29,
                                "sum_other_doc_count": 1226,
                            },
                        },
                    },
                ],
                "doc_count_error_upper_bound": 29,
                "sum_other_doc_count": 1226,
            },
        },
    }

    tree = ResultTree({})
    assert tree._is_nested_node(node)
    assert tree._is_nested_node(node['products'])
    assert not tree._is_nested_node(node['products']['doc_count'])
    assert not tree._is_nested_node(node['products']['product'])
    assert not tree._is_nested_node(node['products']['product']['buckets'])
    assert not tree._is_nested_node(node['products']['product']['doc_count_error_upper_bound'])
    assert not tree._is_nested_node(node['products']['product']['sum_other_doc_count'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]['doc_count'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]['key'])
    assert tree._is_nested_node(node['products']['product']['buckets'][0]['parts'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['doc_count'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['buckets'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['doc_count_error_upper_bound'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['sum_other_doc_count'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['buckets'][0])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['buckets'][0]['doc_count'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['buckets'][0]['key'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['buckets'][0]['avg_part_price'])
    assert not tree._is_nested_node(node['products']['product']['buckets'][0]\
                ['parts']['part']['buckets'][0]['avg_part_price']['value'])


def test_is_nested_node_2():
    node = {
        'products': {
            'doc_count': 1540,
            'parts': {
                'doc_count': 8497,
                'part': {
                    'buckets': [
                        {
                            'avg_part_price': {
                                'value': 19.4,
                            },
                            'doc_count': 907,
                            'key': 'part_5',
                        },
                    ],
                    'doc_count_error_upper_bound': 0,
                    'sum_other_doc_count': 0,
                },
            },
        },
    }

    tree = ResultTree({})
    assert tree._is_nested_node(node)
    assert tree._is_nested_node(node['products'])
    assert not tree._is_nested_node(node['products']['doc_count'])
    assert tree._is_nested_node(node['products']['parts'])
    assert not tree._is_nested_node(node['products']['parts']['doc_count'])
    assert not tree._is_nested_node(node['products']['parts']['part'])
    assert not tree._is_nested_node(node['products']['parts']['part']['buckets'])
    assert not tree._is_nested_node(node['products']['parts']['part']['doc_count_error_upper_bound'])
    assert not tree._is_nested_node(node['products']['parts']['part']['sum_other_doc_count'])
    assert not tree._is_nested_node(node['products']['parts']['part']['buckets'][0])
    assert not tree._is_nested_node(node['products']['parts']['part']['buckets'][0]\
                ['doc_count'])
    assert not tree._is_nested_node(node['products']['parts']['part']['buckets'][0]\
                ['key'])
    assert not tree._is_nested_node(node['products']['parts']['part']['buckets'][0]\
                ['avg_part_price'])
    assert not tree._is_nested_node(node['products']['parts']['part']['buckets'][0]\
                ['avg_part_price']['value'])


def test_remove_nested_aggregations():
    node = {
        "products": {
            "doc_count": 1540,
            "product_type": {
                "buckets": [
                    {
                        "avg_product_price": {
                            "value": 179.53889943074003
                        },
                        "doc_count": 527,
                        "key": "product_type_3"
                    },
                ],
                "doc_count_error_upper_bound": 0,
                "sum_other_doc_count": 0
            },
        },
    }

    expected = {
        "doc_count": 1540,
        "product_type": {
            "buckets": [
                {
                    "avg_product_price": {
                        "value": 179.53889943074003
                    },
                    "doc_count": 527,
                    "key": "product_type_3"
                },
            ],
            "doc_count_error_upper_bound": 0,
            "sum_other_doc_count": 0
        },
    }

    result = ResultTree({})._remove_nested_aggregations(node)

    assert expected == result


def test_remove_nested_aggregations_2():
    node = {
        'products': {
            'doc_count': 100,
            'product': {
                'buckets': [
                    {
                        "doc_count": 38,
                        "key": "product_JQ0JVW20PV",
                        "parts": {
                            "doc_count": 100,
                            "part": {
                                "buckets": [
                                    {
                                        "doc_count": 10,
                                        "key": "part_10",
                                        "avg_part_price": {
                                            "value": 10.11
                                        },
                                    },
                                ],
                                "doc_count_error_upper_bound": 29,
                                "sum_other_doc_count": 1226,
                            },
                        },
                    },
                ],
                "doc_count_error_upper_bound": 29,
                "sum_other_doc_count": 1226,
            },
        },
    }

    expected = {
        'doc_count': 100,
        'product': {
            'buckets': [
                {
                    "key": "product_JQ0JVW20PV",
                    "doc_count": 38,
                    "part": {
                        "buckets": [
                            {
                                "doc_count": 10,
                                "key": "part_10",
                                "avg_part_price": {
                                    "value": 10.11
                                },
                            },
                        ],
                        "doc_count_error_upper_bound": 29,
                        "sum_other_doc_count": 1226,
                    },
                },
            ],
            "doc_count_error_upper_bound": 29,
            "sum_other_doc_count": 1226,
        },
    }

    result = ResultTree({})._remove_nested_aggregations(node)

    assert expected == result


def test_remove_nested_aggregations_3():
    node = {
        'products': {
            'doc_count': 1540,
            'parts': {
                'doc_count': 8497,
                'part': {
                    'buckets': [
                        {
                            'avg_part_price': {
                                'value': 19.4,
                            },
                            'doc_count': 907,
                            'key': 'part_5',
                        },
                    ],
                    'doc_count_error_upper_bound': 0,
                    'sum_other_doc_count': 0,
                },
            },
        },
    }

    expected = {
        'doc_count': 1540,
        'part': {
            'buckets': [
                {
                    'avg_part_price': {
                        'value': 19.4,
                    },
                    'doc_count': 907,
                    'key': 'part_5',
                },
            ],
            'doc_count_error_upper_bound': 0,
            'sum_other_doc_count': 0,
        },
    }


    result = ResultTree({})._remove_nested_aggregations(node)

    assert expected == result


def test_avg_product_price_by_product_type():
    lines = flatten_result(load_output('avg_product_price_by_product_type'))

    assert len(lines) == 5  # One for each product type

    # Lines are sorted by doc_count
    assert lines == list(sorted(lines, key=(lambda l: l['doc_count']), reverse=True))

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metric are present
        assert 'product_type' in line
        assert type(line['product_type']) == six.text_type
        assert 'avg_product_price' in line
        assert type(line['avg_product_price']) == float


def test_avg_part_price_by_part():
    lines = flatten_result(load_output('avg_part_price_by_part'))

    assert len(lines) == 10  # One for each part

    # Lines are sorted by doc_count
    assert lines == list(sorted(lines, key=(lambda l: l['doc_count']), reverse=True))

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metric are present
        assert 'part_id' in line
        assert type(line['part_id']) == six.text_type
        assert 'avg_part_price' in line
        assert type(line['avg_part_price']) == float


def test_avg_part_price_by_product():
    lines = flatten_result(load_output('avg_part_price_by_product'))

    assert len(lines) == 10  # Product agg reached the default 10 limit

    # Lines are sorted by doc_count
    assert lines == list(sorted(lines, key=(lambda l: l['doc_count']), reverse=True))

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metric are present
        assert 'product_id' in line
        assert type(line['product_id']) == six.text_type
        assert 'avg_part_price' in line
        assert type(line['avg_part_price']) == float


def test_avg_part_price_by_product_by_part():
    lines = flatten_result(load_output('avg_part_price_by_product_by_part'))

    assert len(lines) == 10 * 10  # Product agg reached the default 10 limit and 10 parts

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Both aggregations and metric are present
        assert 'product_id' in line
        assert type(line['product_id']) == six.text_type
        assert 'part_id' in line
        assert type(line['part_id']) == six.text_type
        assert 'avg_part_price' in line
        assert type(line['avg_part_price']) == float


def test_avg_product_price_by_shop_by_product_type():
    lines = flatten_result(load_output('avg_product_price_by_shop_by_product_type'))

    assert len(lines) == 5 * 10  # 5 product types and 10 shops

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Both aggregations and metric are present
        assert 'shop_id' in line
        assert type(line['shop_id']) == int
        assert 'product_type' in line
        assert type(line['product_type']) == six.text_type
        assert 'avg_product_price' in line
        assert type(line['avg_product_price']) == float


def test_avg_part_price_by_product_and_by_part():
    lines = flatten_result(load_output('avg_part_price_by_product_and_by_part'))

    assert len(lines) == 10 + 10  # Product agg reached the default 10 limit and 10 parts

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Metric is present
        assert 'avg_part_price' in line
        assert type(line['avg_part_price']) == float
        # Either product aggregation or part aggregation is present
        assert ('product_id' in line and not 'part_id' in line)\
            or ('part_id' in line and 'product_id' not in line)
        if 'product_id' in line:
            assert type(line['product_id']) == six.text_type
        elif 'part_id' in line:
            assert type(line['part_id']) == six.text_type

    # 10 payment lines, sorted by doc_count
    product_lines = [l for l in lines if 'product_id' in l]
    assert len(product_lines) == 10
    assert sorted(product_lines, key=lambda l: l['doc_count'], reverse=True) == product_lines
    # 10 part lines, sorted by doc_count
    part_lines = [l for l in lines if 'part_id' in l]
    assert len(part_lines) == 10
    assert sorted(part_lines, key=lambda l: l['doc_count'], reverse=True) == part_lines



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

import pytest

from fiqs.aggregations import Avg, Sum, DateHistogram, Count
from fiqs.query import FQuery
from fiqs.testing.models import Sale, TrafficCount
from fiqs.testing.utils import get_search
from fiqs.tests.conftest import write_output, write_fquery_output


@pytest.mark.docker
def test_count(elasticsearch_sale):
    assert get_search().count() == 500


@pytest.mark.xfail  # See https://github.com/elastic/elasticsearch/issues/23776
@pytest.mark.docker
def test_offset_date_histogram(elasticsearch_sale):
    start = datetime(2016, 2, 1, 6)
    end = start + timedelta(days=2, hours=2)

    search = get_search()
    search = search.filter('range', **{
        'timestamp': {
            'gte': start.isoformat(),
            'lte': end.isoformat(),
        },
    })

    fquery = FQuery(search).values(
        Count(Sale),
    ).group_by(
        DateHistogram(
            Sale.timestamp,
            interval='1d',
            min=start,
            max=end,
            offset='+6h',
        )
    )

    result = fquery.eval(flat=False)
    aggregations = result._d_['aggregations']
    buckets = aggregations['timestamp']['buckets']

    keys = [datetime.utcfromtimestamp(bucket['key'] / 1000) for bucket in buckets]
    expected_keys = [
        start,
        start + timedelta(days=1),
        start + timedelta(days=2),
    ]
    assert keys == expected_keys


@pytest.mark.docker
def test_write_traffic_search_outputs(elasticsearch_traffic):
    # Total in traffic and out traffic
    write_fquery_output(
        FQuery(get_search()).values(
            Sum(TrafficCount.incoming_traffic),
            Sum(TrafficCount.outgoing_traffic),
        ),
        'total_in_traffic_and_total_out_traffic',
    )

    # Total in traffic and out traffic by shop id
    write_fquery_output(
        FQuery(get_search()).values(
            Sum(TrafficCount.incoming_traffic),
            Sum(TrafficCount.outgoing_traffic),
        ).group_by(
            TrafficCount.shop_id,
        ),
        'total_in_traffic_and_total_out_traffic_by_shop',
    )


@pytest.mark.docker
def test_write_nested_search_output(elasticsearch_sale):
    # Average product price by product type
    write_fquery_output(
        FQuery(get_search()).values(
            avg_product_price=Avg(Sale.product_price),
        ).group_by(
            Sale.product_type,
        ),
        'avg_product_price_by_product_type',
    )

    # Average part price by part
    write_fquery_output(
        FQuery(get_search()).values(
            avg_part_price=Avg(Sale.part_price),
        ).group_by(
            Sale.part_id,
        ),
        'avg_part_price_by_part',
    )

    # Average part price by product
    write_fquery_output(
        FQuery(get_search()).values(
            avg_part_price=Avg(Sale.part_price),
        ).group_by(
            Sale.product_id,
            Sale.parts,
        ),
        'avg_part_price_by_product',
    )

    # Average part price by product by part
    write_fquery_output(
        FQuery(get_search()).values(
            avg_part_price=Avg(Sale.part_price),
        ).group_by(
            Sale.product_id,
            Sale.part_id,
        ),
        'avg_part_price_by_product_by_part',
    )

    # Average product price by shop by product
    write_fquery_output(
        FQuery(get_search()).values(
            avg_product_price=Avg(Sale.product_price),
        ).group_by(
            Sale.shop_id,
            Sale.product_type,
        ),
        'avg_product_price_by_shop_by_product_type',
    )

    # Average part price by product and by part
    # This type of query is not possible with FQuery
    search = get_search()
    products_bucket = search.aggs.bucket(
        'products', 'nested', path='products',
    )
    products_bucket.bucket(
        'product_id', 'terms', field='products.product_id',
    ).bucket(
        'parts', 'nested', path='products.parts',
    ).metric(
        'avg_part_price', 'avg', field='products.parts.part_price',
    )
    products_bucket.bucket(
        'parts', 'nested', path='products.parts',
    ).bucket(
        'part_id', 'terms', field='products.parts.part_id',
    ).metric(
        'avg_part_price', 'avg', field='products.parts.part_price',
    )
    write_output(search, 'avg_part_price_by_product_and_by_part')


@pytest.mark.docker
def test_write_search_outputs(elasticsearch_sale):
    # Nothing :o
    write_output(get_search(), 'no_aggregate_no_metric')

    # Total sales by shop
    write_fquery_output(
        FQuery(get_search()).values(
            total_sales=Sum(Sale.price),
        ).group_by(
            Sale.shop_id,
        ),
        'total_sales_by_shop',
    )

    # Total sales by payment type
    write_fquery_output(
        FQuery(get_search()).values(
            total_sales=Sum(Sale.price),
        ).group_by(
            Sale.payment_type,
        ),
        'total_sales_by_payment_type',
    )

    # Total sales by shop by payment type
    write_fquery_output(
        FQuery(get_search()).values(
            total_sales=Sum(Sale.price),
        ).group_by(
            Sale.payment_type,
            Sale.shop_id,
        ),
        'total_sales_by_payment_type_by_shop',
    )

    # Total sales day by day
    write_fquery_output(
        FQuery(get_search()).values(
            total_sales=Sum(Sale.price),
        ).group_by(
            DateHistogram(
                Sale.timestamp,
                interval='1d',
            ),
        ),
        'total_sales_day_by_day',
    )

    # Number of sales by shop
    write_fquery_output(
        FQuery(get_search()).values(
            Count(Sale.id),
        ).group_by(
            Sale.shop_id,
        ),
        'nb_sales_by_shop',
    )

    # Total sales day by day, by shop and by payment type
    # This type of query is not possible with FQuery
    search = get_search()
    agg = search.aggs.bucket(
        'timestamp', 'date_histogram', field='timestamp', interval='1d',
    )
    agg.bucket(
        'shop_id', 'terms', field='shop_id',
    ).metric(
        'total_sales', 'sum', field='price',
    )
    agg.bucket(
        'payment_type', 'terms', field='payment_type',
    ).metric(
        'total_sales', 'sum', field='price',
    )
    write_output(search, 'total_sales_day_by_day_by_shop_and_by_payment')

    # Total sales and average sales by shop
    write_fquery_output(
        FQuery(get_search()).values(
            total_sales=Sum(Sale.price),
            avg_sales=Avg(Sale.price),
        ).group_by(
            Sale.shop_id,
        ),
        'total_and_avg_sales_by_shop',
    )

    # Total sales, no aggregations
    write_fquery_output(
        FQuery(get_search()).values(
            total_sales=Sum(Sale.price),
        ),
        'total_sales',
    )

    # Total sales and avg sales, no aggregations
    write_fquery_output(
        FQuery(get_search()).values(
            total_sales=Sum(Sale.price),
            avg_sales=Avg(Sale.price),
        ),
        'total_sales_and_avg_sales',
    )

    # Total sales by shop and by payment type
    # This type of query is not possible with FQuery
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id',
    ).metric(
        'total_sales', 'sum', field='price',
    )
    search.aggs.bucket(
        'payment_type', 'terms', field='payment_type',
    ).metric(
        'total_sales', 'sum', field='price',
    )
    write_output(search, 'total_sales_by_shop_and_by_payment')



# -*- coding: utf-8 -*-

import json
import random

from fiqs.testing.gen_data import random_string, random_timestamp, random_shop_id


PRODUCT_TYPES = ['product_type_{}'.format(i) for i in xrange(5)]
PRODUCT_IDS = ['product_{}'.format(random_string(10)) for _ in xrange(50)]
PRODUCTS = [
    (product_id, random.choice(PRODUCT_TYPES))
    for product_id in PRODUCT_IDS
]


def gen_parts(product_price):
    parts = []
    nb_parts = random.randint(1, 10)

    for i in xrange(nb_parts):
        parts.append({
            'part_id': 'part_{}'.format(random.randint(1, 10)),
            'warehouse_id': 'warehouse_{}'.format(random.randint(1, 10)),
            'part_price': int(product_price / (nb_parts * 1.5)),
        })

    return parts


def gen_products(price):
    products = []
    nb_products = random.randint(1, 5)

    for i in xrange(nb_products):
        if i == nb_products - 1:
            product_price = price - sum(p['product_price'] for p in products)
        else:
            product_price = price / 10

        product_id, product_type = random.choice(PRODUCTS)

        products.append({
            'product_id': product_id,
            'product_type': product_type,
            'product_price': product_price,
            'parts': gen_parts(product_price),
        })

    return products


def gen_shop_data(size):
    for i in xrange(size):
        price = random.randint(10, 1000)
        print(json.dumps({
            'id': i + 1,
            'shop_id': random_shop_id(),
            'client_id': 'client_{}'.format(random_string(10)),
            'timestamp': random_timestamp(),
            'price': price,
            'payment_type': random.choice(['wire_transfer', 'cash', 'store_credit', ]),
            'products': gen_products(price),
        }))


if __name__ == '__main__':
    size = 500
    gen_shop_data(size)



# -*- coding: utf-8 -*-

import random
import string
import time
from datetime import datetime

from faker import Factory


fake = Factory.create()


def random_timestamp():
    start = datetime(2016, 1, 1)
    end = datetime(2016, 1, 31)

    return int(1000 * time.mktime(fake.date_time_between_dates(start, end).timetuple()))


def random_string(length=10):
    return ''.join(random.choice(string.ascii_uppercase + string.digits + ' ') for _ in range(length))


def random_shop_id():
    return random.randint(1, 10)



# -*- coding: utf-8 -*-

from collections import Counter
from datetime import datetime

import pytest

from fiqs import fields
from fiqs.aggregations import Sum, Count, Avg, DateHistogram, Addition, Ratio, Subtraction
from fiqs.exceptions import ConfigurationError
from fiqs.fields import DataExtendedField, FieldWithChoices
from fiqs.models import Model
from fiqs.query import FQuery

from fiqs.testing.models import Sale, TrafficCount
from fiqs.testing.utils import get_search
from fiqs.tests.conftest import load_output


def test_one_metric():
    expected = get_search().aggs.metric(
        'total_sales', 'sum', field='price',
    ).to_dict()

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    )
    search = fquery._configure_search()

    assert expected == search.to_dict()


def test_one_aggregation():
    expected = get_search().aggs.metric(
        'shop_id', 'terms', field='shop_id',
    ).to_dict()

    fquery = FQuery(get_search()).values(
        Count(Sale),
    ).group_by(
        Sale.shop_id,
    )
    search = fquery._configure_search()

    assert expected == search.to_dict()


def test_one_aggregation_one_metric():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id',
    ).metric(
        'total_sales', 'sum', field='price',
    )

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_one_aggregation_count():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id',
    )

    fquery = FQuery(get_search()).values(
        Count(Sale),
    ).group_by(
        Sale.shop_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_two_aggregations_one_metric():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id',
    ).bucket(
        'client_id', 'terms', field='client_id',
    ).metric(
        'total_sales', 'sum', field='price',
    )

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
        Sale.client_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_one_aggregation_two_metrics():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id',
    ).metric(
        'total_sales', 'sum', field='price',
    ).metric(
        'avg_sales', 'avg', field='price',
    )

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
        avg_sales=Avg(Sale.price),
    ).group_by(
        Sale.shop_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_two_aggregations_two_metrics():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id',
    ).bucket(
        'client_id', 'terms', field='client_id',
    ).metric(
        'total_sales', 'sum', field='price',
    ).metric(
        'avg_sales', 'avg', field='price',
    )

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
        avg_sales=Avg(Sale.price),
    ).group_by(
        Sale.shop_id,
        Sale.client_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_date_histogram():
    start = datetime(2016, 1, 1)
    end = datetime(2016, 1, 31)

    date_histogram_params = {
        'field': 'timestamp',
        'interval': '1d',
        'min_doc_count': 0,
        'extended_bounds': {
            'min': start,
            'max': end,
        },
    }

    search = get_search()
    search.aggs.bucket(
        'timestamp', 'date_histogram', **date_histogram_params
    ).metric(
        'total_sales', 'sum', field='price',
    )

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        DateHistogram(
            Sale.timestamp,
            interval='1d',
            min=start,
            max=end,
        ),
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_one_nested_aggregation_one_metric():
    search = get_search()
    search.aggs.bucket(
        'products', 'nested', path='products',
    ).bucket(
        'product_type', 'terms', field='products.product_type',
    ).metric(
        'avg_product_price', 'avg', field='products.product_price',
    )

    fquery = FQuery(get_search()).values(
        avg_product_price=Avg(Sale.product_price),
    ).group_by(
        Sale.products,
        Sale.product_type,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_two_nested_aggregations_one_metric():
    search = get_search()
    search.aggs.bucket(
        'products', 'nested', path='products',
    ).bucket(
        'product_id', 'terms', field='products.product_id',
    ).bucket(
        'parts', 'nested', path='products.parts',
    ).bucket(
        'part_id', 'terms', field='products.parts.part_id',
    ).metric(
        'avg_part_price', 'avg', field='products.parts.part_price',
    )

    fquery = FQuery(get_search()).values(
        avg_part_price=Avg(Sale.part_price),
    ).group_by(
        Sale.products,
        Sale.product_id,
        Sale.parts,
        Sale.part_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_nested_parent_automatically_added():
    """Need to add one parent for each group_by key"""
    search = get_search()
    search.aggs.bucket(
        'products', 'nested', path='products',
    ).bucket(
        'product_id', 'terms', field='products.product_id',
    ).bucket(
        'parts', 'nested', path='products.parts',
    ).bucket(
        'part_id', 'terms', field='products.parts.part_id',
    ).metric(
        'avg_part_price', 'avg', field='products.parts.part_price',
    )

    fquery = FQuery(get_search()).values(
        avg_part_price=Avg(Sale.part_price),
    ).group_by(
        Sale.product_id,
        Sale.part_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_nested_parent_automatically_added_2():
    """Need to add multiple parents for one group_by key"""
    search = get_search()
    search.aggs.bucket(
        'products', 'nested', path='products',
    ).bucket(
        'parts', 'nested', path='products.parts',
    ).bucket(
        'part_id', 'terms', field='products.parts.part_id',
    ).metric(
        'avg_part_price', 'avg', field='products.parts.part_price',
    )

    fquery = FQuery(get_search()).values(
        avg_part_price=Avg(Sale.part_price),
    ).group_by(
        Sale.part_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_nested_parent_automatically_added_3():
    """Should this work differently? Can it?"""
    search = get_search()
    search.aggs.bucket(
        'products', 'nested', path='products',
    ).bucket(
        'product_id', 'terms', field='products.product_id',
    ).bucket(
        'parts', 'nested', path='products.parts',
    ).metric(
        'avg_part_price', 'avg', field='products.parts.part_price',
    )

    fquery = FQuery(get_search()).values(
        avg_part_price=Avg(Sale.part_price),
    ).group_by(
        Sale.product_id,
        Sale.parts,  # Can we get rid of this?
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_default_size():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id', size=5,
    ).bucket(
        'client_id', 'terms', field='client_id', size=5,
    ).metric(
        'total_sales', 'sum', field='price',
    )

    fquery = FQuery(get_search(), default_size=5).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
        Sale.client_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_aggregation_size():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id', size=5,
    ).bucket(
        'client_id', 'terms', field='client_id',
    ).metric(
        'total_sales', 'sum', field='price',
    )

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        DataExtendedField(Sale.shop_id, size=5),
        Sale.client_id,
    )
    fsearch = fquery._configure_search()

    assert search.to_dict() == fsearch.to_dict()


def test_order_by():
    expected = {
        'aggs': {
            'shop_id': {
                'aggs': {
                    'total_sales': {'sum': {'field': 'price'}},
                },
                'terms': {
                    'field': 'shop_id',
                    'order': {'total_sales': 'desc'},
                }
            }
        },
        'query': {'match_all': {}},
    }

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
    ).order_by(
        {'total_sales': 'desc'}
    )
    fsearch = fquery._configure_search()

    assert fsearch.to_dict() == expected


def test_order_by_multiple_group_by():
    expected = {
        'aggs': {
            'shop_id': {
                'aggs': {
                    'client_id': {
                        'aggs': {
                            'total_sales': {'sum': {'field': 'price'}},
                        },
                        'terms': {
                            'field': 'client_id',
                            'order': {'total_sales': 'desc'},
                        },
                    },
                },
                'terms': {'field': 'shop_id'},
            },
        },
        'query': {'match_all': {}}
    }

    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
        Sale.client_id,
    ).order_by(
        {'total_sales': 'desc'}
    )
    fsearch = fquery._configure_search()

    assert fsearch.to_dict() == expected


def test_order_by_count():
    pass


def test_computed_automatically_added():
    search = get_search()
    search.aggs.bucket(
        'shop_id', 'terms', field='shop_id',
    ).metric(
        str(Sum(TrafficCount.incoming_traffic)), 'sum', field='incoming_traffic',
    ).metric(
        str(Sum(TrafficCount.outgoing_traffic)), 'sum', field='outgoing_traffic',
    )

    fquery = FQuery(get_search()).values(
        Addition(
            Sum(TrafficCount.incoming_traffic),
            Sum(TrafficCount.outgoing_traffic),
        ),
        Sum(TrafficCount.incoming_traffic),
        Sum(TrafficCount.outgoing_traffic),
    ).group_by(
        TrafficCount.shop_id,
    )
    fsearch = fquery._configure_search()

    assert fsearch.to_dict() == search.to_dict()

    fquery = FQuery(get_search()).values(
        Addition(
            Sum(TrafficCount.incoming_traffic),
            Sum(TrafficCount.outgoing_traffic),
        ),
    ).group_by(
        TrafficCount.shop_id,
    )
    fsearch = fquery._configure_search()

    assert fsearch.to_dict() == search.to_dict()

    fquery = FQuery(get_search()).values(
        Ratio(
            Sum(TrafficCount.incoming_traffic),
            Addition(
                Sum(TrafficCount.incoming_traffic),
                Sum(TrafficCount.outgoing_traffic),
            ),
        ),
    ).group_by(
        TrafficCount.shop_id,
    )
    fsearch = fquery._configure_search()

    assert fsearch.to_dict() == search.to_dict()

###################
# Flatten results #
###################

def test_flatten_result_count():
    search = get_search()
    fquery = FQuery(search).values(
        Count(Sale),
    ).group_by(
        Sale.shop_id,
    )

    result = load_output('nb_sales_by_shop')
    lines = fquery._flatten_result(result)

    assert len(lines) == 10  # One for each shop

    # Lines are sorted by doc_count
    assert lines == sorted(lines, key=(lambda l: l['doc_count']), reverse=True)

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation is present
        assert 'shop_id' in line
        assert type(line['shop_id']) == int


def test_flatten_result_cast_sum_to_int():
    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
    )

    result = load_output('total_sales_by_shop')
    lines = fquery._flatten_result(result)

    assert len(lines) == 10  # One for each shop

    # Lines are sorted by doc_count
    assert lines == sorted(lines, key=(lambda l: l['doc_count']), reverse=True)

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metric are present
        assert 'shop_id' in line
        assert type(line['shop_id']) == int
        # Total sales aggregation results were casted to int
        assert 'total_sales' in line
        assert type(line['total_sales']) == int


def test_flatten_result_cast_timestamp():
    fquery = FQuery(get_search()).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        DateHistogram(
            Sale.timestamp,
            interval='1d',
        ),
    )

    result = load_output('total_sales_day_by_day')
    lines = fquery._flatten_result(result)

    assert len(lines) == 31

    for line in lines:
        # Doc count is present
        assert 'doc_count' in line
        assert type(line['doc_count']) == int
        # Aggregation and metric are present
        assert 'timestamp' in line
        assert type(line['timestamp']) == datetime
        # Total sales aggregation results were casted to int
        assert 'total_sales' in line
        assert type(line['total_sales']) == int


def test_computed_field():
    computed_field = Addition(
        Sum(TrafficCount.incoming_traffic),
        Sum(TrafficCount.outgoing_traffic),
    )

    fquery = FQuery(get_search()).values(
        computed_field,
    ).group_by(
        TrafficCount.shop_id,
    )

    result = load_output('total_in_traffic_and_total_out_traffic')
    lines = fquery._flatten_result(result)

    assert len(lines) == 1

    line = lines[0]
    # Base expressions are there
    assert str(Sum(TrafficCount.incoming_traffic)) in line
    assert str(Sum(TrafficCount.outgoing_traffic)) in line
    # Computed field is present
    assert str(computed_field) in line


def test_multiple_computed_fields():
    addition = Addition(
        Sum(TrafficCount.incoming_traffic),
        Sum(TrafficCount.outgoing_traffic),
    )
    computed_field = Ratio(
        Sum(TrafficCount.incoming_traffic),
        addition,
    )

    fquery = FQuery(get_search()).values(
        computed_field,
    ).group_by(
        TrafficCount.shop_id,
    )

    result = load_output('total_in_traffic_and_total_out_traffic')
    lines = fquery._flatten_result(result)

    assert len(lines) == 1

    line = lines[0]
    # Base expressions are there
    assert str(Sum(TrafficCount.incoming_traffic)) in line
    assert str(Sum(TrafficCount.outgoing_traffic)) in line
    # Addition is there
    assert str(addition) in line
    # Computed field is present
    assert str(computed_field) in line


def test_multiple_computed_fields_2():
    addition = Addition(
        Sum(TrafficCount.incoming_traffic),
        Sum(TrafficCount.outgoing_traffic),
    )
    subtraction = Subtraction(
        Sum(TrafficCount.incoming_traffic),
        Sum(TrafficCount.outgoing_traffic),
    )
    computed_field = Ratio(
        addition,
        subtraction,
    )

    fquery = FQuery(get_search()).values(
        computed_field,
    ).group_by(
        TrafficCount.shop_id,
    )

    result = load_output('total_in_traffic_and_total_out_traffic')
    lines = fquery._flatten_result(result)

    assert len(lines) == 1

    line = lines[0]
    # Base expressions are there
    assert str(Sum(TrafficCount.incoming_traffic)) in line
    assert str(Sum(TrafficCount.outgoing_traffic)) in line
    # Addition and subtraction are there
    assert str(addition) in line
    assert str(subtraction)
    # Computed field is present
    assert str(computed_field) in line


def test_computed_fields_raise_when_non_flat():
    addition = Addition(
        Sum(TrafficCount.incoming_traffic),
        Sum(TrafficCount.outgoing_traffic),
    )

    fquery = FQuery(get_search()).values(
        addition,
    ).group_by(
        TrafficCount.shop_id,
    )
    with pytest.raises(ConfigurationError):
        fquery.eval(flat=False)

########################
# Fill missing buckets #
########################

def test_fill_missing_buckets_nothing_to_do():
    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
    )

    result = load_output('total_sales_by_shop')
    lines = fquery._flatten_result(result)
    assert lines == fquery._add_missing_lines(result, lines)


def test_fill_missing_buckets_cannot_do_anything():
    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.shop_id,
    )

    result = load_output('total_sales_by_shop')
    result['aggregations']['shop_id']['buckets'] = [
        bucket for bucket in result['aggregations']['shop_id']['buckets']
        if bucket['key'] != 1
    ]

    lines = fquery._flatten_result(result)
    assert lines == fquery._add_missing_lines(result, lines)

    assert len(lines) == 9


def test_fill_missing_buckets_custom_choices():
    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        FieldWithChoices(Sale.shop_id, choices=range(1, 11)),
    )

    result = load_output('total_sales_by_shop')
    result['aggregations']['shop_id']['buckets'] = [
        bucket for bucket in result['aggregations']['shop_id']['buckets']
        if bucket['key'] != 1
    ]

    lines = fquery._flatten_result(result)
    assert len(lines) == 9
    assert sorted([line['shop_id'] for line in lines]) == list(range(2, 11))

    lines == fquery._add_missing_lines(result, lines)
    assert len(lines) == 10
    assert sorted([line['shop_id'] for line in lines]) == list(range(1, 11))

    added_line = [line for line in lines if line['shop_id'] == 1][0]
    assert added_line == {
        'shop_id': 1,
        'total_sales': None,
        'doc_count': 0,
    }


def test_fill_missing_buckets_field_choices():
    class SaleWithChoices(Model):
        shop_id = fields.IntegerField(choices=range(1, 11))
        price = fields.IntegerField()

    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(SaleWithChoices.price),
    ).group_by(
        SaleWithChoices.shop_id,
    )

    result = load_output('total_sales_by_shop')
    result['aggregations']['shop_id']['buckets'] = [
        bucket for bucket in result['aggregations']['shop_id']['buckets']
        if bucket['key'] != 1
    ]

    lines = fquery._flatten_result(result)
    assert len(lines) == 9
    assert sorted([line['shop_id'] for line in lines]) == list(range(2, 11))

    lines == fquery._add_missing_lines(result, lines)
    assert len(lines) == 10
    assert sorted([line['shop_id'] for line in lines]) == list(range(1, 11))

    added_line = [line for line in lines if line['shop_id'] == 1][0]
    assert added_line == {
        'shop_id': 1,
        'total_sales': None,
        'doc_count': 0,
    }


def test_fill_missing_buckets_field_choices_pretty():
    pretty_choices = [(i, 'Shop {}'.format(i)) for i in range(1, 11)]

    class SaleWithPrettyChoices(Model):
        shop_id = fields.IntegerField(choices=pretty_choices)
        price = fields.IntegerField()

    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(SaleWithPrettyChoices.price),
    ).group_by(
        SaleWithPrettyChoices.shop_id,
    )

    result = load_output('total_sales_by_shop')
    result['aggregations']['shop_id']['buckets'] = [
        bucket for bucket in result['aggregations']['shop_id']['buckets']
        if bucket['key'] != 1
    ]

    lines = fquery._flatten_result(result)
    assert len(lines) == 9
    assert sorted([line['shop_id'] for line in lines]) == list(range(2, 11))

    lines == fquery._add_missing_lines(result, lines)
    assert len(lines) == 10
    assert sorted([line['shop_id'] for line in lines]) == list(range(1, 11))

    added_line = [line for line in lines if line['shop_id'] == 1][0]
    assert added_line == {
        'shop_id': 1,
        'total_sales': None,
        'doc_count': 0,
    }


def test_fill_missing_buckets_values_in_other_agg():
    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        Sale.payment_type,
        Sale.shop_id,
    )

    result = load_output('total_sales_by_payment_type_by_shop_id')
    wire_transfer_bucket = [
        b for b in result['aggregations']['payment_type']['buckets']
        if b['key'] == 'wire_transfer'
    ][0]
    wire_transfer_bucket['shop_id']['buckets'] = [
        b for b in wire_transfer_bucket['shop_id']['buckets']
        if b['key'] != 1
    ]

    lines = fquery._flatten_result(result)
    assert len(lines) == 29
    # All shop ids are present three times
    counter = Counter([line['shop_id'] for line in lines])
    for i in range(2, 11):
        assert counter[i] == 3
    # Except the one I removed
    assert counter[1] == 2

    lines == fquery._add_missing_lines(result, lines)
    assert len(lines) == 30
    # All shop ids are present three times
    counter = Counter([line['shop_id'] for line in lines])
    for i in range(1, 11):
        assert counter[i] == 3

    added_line = [
        line for line in lines
        if line['shop_id'] == 1 and line['payment_type'] == 'wire_transfer'
    ][0]
    assert added_line == {
        'shop_id': 1,
        'total_sales': None,
        'doc_count': 0,
        'payment_type': 'wire_transfer',
    }


def test_fill_missing_buckets_range_nothing_to_do():
    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        DateHistogram(
            Sale.timestamp,
            interval='1d',
            min=datetime(2016, 1, 1),
            max=datetime(2016, 1, 31),
        ),
    )
    fquery._configure_search()

    result = load_output('total_sales_day_by_day')

    lines = fquery._flatten_result(result)
    assert lines == fquery._add_missing_lines(result, lines)
    assert len(lines) == 31


def test_fill_missing_buckets_date_histogram():
    search = get_search()
    fquery = FQuery(search).values(
        total_sales=Sum(Sale.price),
    ).group_by(
        DateHistogram(
            Sale.timestamp,
            interval='1d',
            min=datetime(2015, 12, 1),
            max=datetime(2016, 1, 31),
        ),
    )
    fquery._configure_search()

    result = load_output('total_sales_day_by_day')

    lines = fquery._flatten_result(result)
    lines == fquery._add_missing_lines(result, lines)
    assert len(lines) == 62



# -*- coding: utf-8 -*-

import json
import os
import time

from elasticsearch.helpers import bulk

from elasticsearch_dsl import Mapping, Nested

import pytest

from fiqs.testing.utils import get_client


SALE_INDEX_NAME = 'test_sale'
TRAFFIC_INDEX_NAME = 'test_traffic'

SALE_FIXTURE_PATH = 'fiqs/testing/fixtures/shop_fixture.json'
TRAFFIC_FIXTURE_PATH = 'fiqs/testing/fixtures/traffic_fixture.json'

SALE_DOC_TYPE = 'sale'
TRAFFIC_DOC_TYPE = 'traffic_count'


def sale_mapping():
    m = Mapping(SALE_DOC_TYPE)

    m.meta('dynamic', 'strict')

    m.field('id', 'integer')
    m.field('shop_id', 'integer')
    m.field('client_id', 'keyword')
    m.field('timestamp', 'date')
    m.field('price', 'integer')
    m.field('payment_type', 'keyword')

    products = Nested()
    products.field('product_id', 'keyword')
    products.field('product_type', 'keyword')
    products.field('product_price', 'integer')

    parts = Nested()
    parts.field('part_id', 'keyword')
    parts.field('warehouse_id', 'keyword')
    parts.field('part_price', 'integer')

    products.field('parts', parts)
    m.field('products', products)

    return m


def traffic_mapping():
    m = Mapping(TRAFFIC_DOC_TYPE)

    m.meta('dynamic', 'strict')

    m.field('id', 'integer')
    m.field('shop_id', 'integer')
    m.field('timestamp', 'date')
    m.field('duration', 'integer')
    m.field('incoming_traffic', 'integer')
    m.field('outgoing_traffic', 'integer')

    return m


def create_index(client, index_name, mapping):
    request_body = {'mappings': mapping.to_dict()}
    return client.indices.create(index=index_name, body=request_body, ignore=400)


def delete_index(client, index_name):
    return client.indices.delete(index=index_name)


def insert_documents(client, index_name, fixture_path, doc_type):
    with open(fixture_path) as f:
        lines = f.readlines()

    events = [json.loads(line) for line in lines]

    actions = []
    for event in events:
        actions.append({
            '_index': index_name,
            '_type': 'sale',
            '_source': event,
            '_id': event['id'],
        })

    return bulk(client, actions)


def insert_sale_documents(client):
    create_sale_index(client)
    insert_documents(client, SALE_INDEX_NAME, SALE_FIXTURE_PATH, SALE_DOC_TYPE)


def insert_traffic_documents(client):
    create_traffic_index(client)
    insert_documents(client, TRAFFIC_INDEX_NAME, TRAFFIC_FIXTURE_PATH, TRAFFIC_DOC_TYPE)


def create_sale_index(client):
    create_index(client, SALE_INDEX_NAME, sale_mapping())


def create_traffic_index(client):
    create_index(client, TRAFFIC_INDEX_NAME, traffic_mapping())


def delete_sale_index(client):
    delete_index(client, SALE_INDEX_NAME)


def delete_traffic_index(client):
    delete_index(client, TRAFFIC_INDEX_NAME)


@pytest.fixture
def elasticsearch_sale(request):
    client = get_client()
    insert_sale_documents(client)

    request.addfinalizer(lambda: delete_sale_index(client))
    time.sleep(1)

    return client


@pytest.fixture
def elasticsearch_traffic(request):
    client = get_client()
    insert_traffic_documents(client)

    request.addfinalizer(lambda: delete_traffic_index(client))
    time.sleep(1)

    return client


BASE_PATH = 'fiqs/testing/outputs/'


def write_output(search, name):
    result = search.execute()

    path = os.path.join(BASE_PATH, '{}.json'.format(name))
    with open(path, 'w') as f:
        d = result._d_
        d.pop('took', None)  # Not used and may change between calls

        json.dump(result._d_, f, indent=4, ensure_ascii=False, encoding='utf-8', sort_keys=True)


def write_fquery_output(fquery, name):
    result = fquery.eval(flat=False)

    path = os.path.join(BASE_PATH, '{}.json'.format(name))
    with open(path, 'w') as f:
        d = result._d_
        d.pop('took', None)  # Not used and may change between calls

        json.dump(result._d_, f, indent=4, ensure_ascii=False, encoding='utf-8', sort_keys=True)


def load_output(name):
    path = os.path.join(BASE_PATH, '{}.json'.format(name))

    with open(path, 'r') as f:
        output = json.load(f)

    return output



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

from fiqs.aggregations import DateHistogram

from fiqs.testing.models import Sale


def get_date_histogram(**kwargs):
    date_histogram = DateHistogram(
        Sale.timestamp,
        **kwargs
    )
    date_histogram.agg_params()  # To set min and max
    return date_histogram


def test_date_histogram_choice_keys_day_intervals():
    start = datetime(2016, 1, 1, 6)
    end = datetime(2016, 2, 1, 6)
    delta_days = 1
    nb_days = (end - start).days + delta_days

    expected_start = start.replace(hour=0)

    date_histogram = get_date_histogram(min=start, max=end, interval='1d')
    expected_keys = [expected_start + timedelta(days=i) for i in range(0, nb_days, delta_days)]
    keys = date_histogram.choice_keys()
    assert expected_keys == keys

    date_histogram = get_date_histogram(min=start, max=end, interval='day')
    expected_keys = [expected_start + timedelta(days=i) for i in range(0, nb_days, delta_days)]
    keys = date_histogram.choice_keys()
    assert expected_keys == keys


def test_date_histogram_choice_keys_day_intervals_2():
    start = datetime(2016, 1, 1, 6)
    end = datetime(2016, 2, 1, 6)
    delta_days = 2
    nb_days = (end - start).days + delta_days

    expected_start = datetime(2015, 12, 31)

    date_histogram = get_date_histogram(min=start, max=end, interval='2d')
    expected_keys = [expected_start + timedelta(days=i) for i in range(0, nb_days, delta_days)]
    keys = date_histogram.choice_keys()
    assert expected_keys == keys


def test_date_histogram_choice_keys_second_intervals():
    start = datetime(2016, 1, 1, 6, microsecond=123)
    end = start + timedelta(days=1)
    delta_seconds = 433
    nb_seconds = int((end - start).total_seconds()) + delta_seconds

    expected_start = datetime(2016, 1, 1, 5, 55, 37)

    date_histogram = get_date_histogram(min=start, max=end, interval='433s')
    expected_keys = [
        expected_start + timedelta(seconds=i)
        for i in range(0, nb_seconds, delta_seconds)
    ]
    keys = date_histogram.choice_keys()
    assert expected_keys == keys



# -*- coding: utf-8 -*-

from collections import OrderedDict
from itertools import product
from six import text_type as unicode

from fiqs import flatten_result
from fiqs.aggregations import Aggregate
from fiqs.exceptions import ConfigurationError
from fiqs.fields import NestedField, ReverseNestedField, Field


def calc_group_by_keys(group_by_fields, nested=True):
    ret = []
    for field in group_by_fields:
        if isinstance(field, Aggregate):
            ret.append(field.field.key)
        elif isinstance(field, NestedField):
            if nested:
                ret.append(field.key)
        elif isinstance(field, ReverseNestedField):
            continue
        else:
            ret.append(field.key)
    return ret


class FQuery(object):
    def __init__(self, search, default_size=None):
        self.search = search

        if default_size == 0:
            default_size = 2 ** 32 - 1
        self.default_size = default_size

        self._expressions = OrderedDict()
        self._group_by = []
        self._order_by = {}

    def values(self, *expressions, **named_expressions):
        # /!\ named_expressions may not be correctly ordered
        # It may break some Operation fields
        exps = OrderedDict()
        for exp in expressions:
            exps[str(exp)] = exp

        exps.update(named_expressions)
        self._expressions.update(exps)

        self._check_exps_for_computed_are_present()

        return self

    def group_by(self, *args):
        self._group_by += args

        self._check_nested_parents_are_present()

        return self

    def order_by(self, order_dict):
        self._order_by.update(order_dict)

        return self

    def eval(self, flat=True, fill_missing_buckets=True):
        # Raise if computed fields are present, and we are not in flat mode
        if not flat:
            for expression in self._expressions.values():
                if expression.is_computed():
                    raise ConfigurationError(u'Cannot use computed fields in non-flat mode')

        search = self._configure_search()
        result = search.execute()

        if flat:
            lines = self._flatten_result(result)

            if fill_missing_buckets:
                lines = self._add_missing_lines(result, lines)

            return lines
        else:
            return result

    ################
    # Internal API #
    ################
    def _check_exps_for_computed_are_present(self):
        while True:
            exps_to_add = {}
            expression_keys = [str(exp) for exp in self._expressions.values()]

            for key, expression in self._expressions.items():
                if not expression.is_computed():
                    continue

                operands = expression.operands
                for op in operands:
                    if str(op) not in expression_keys:
                        exps_to_add[str(op)] = op
                        expression_keys.append(str(op))

            if not exps_to_add:
                break

            self._expressions.update(exps_to_add)

    def _check_nested_parents_are_present(self):
        while True:
            nested_fields_to_add = {}

            for idx, field in enumerate(self._group_by):
                if not isinstance(field, Field):
                    continue

                parent_field = field.get_parent_field()
                if not parent_field:
                    continue

                if parent_field in self._group_by:
                    continue

                if parent_field in nested_fields_to_add.values():
                    continue

                nested_fields_to_add[idx] = parent_field

            if not nested_fields_to_add:
                break

            indices = sorted(nested_fields_to_add.keys(), reverse=True)
            for idx in indices:
                self._group_by.insert(idx, nested_fields_to_add[idx])

    def _configure_search(self):
        agg = self._configure_aggregations()
        self._configure_values(agg)

        return self.search

    def _configure_aggregations(self):
        current_agg = self.search.aggs
        last_idx = len(self._group_by) - 1

        for idx, field_or_exp in enumerate(self._group_by):
            if isinstance(field_or_exp, Aggregate):
                current_agg = current_agg.bucket(**field_or_exp.agg_params())
                continue

            if isinstance(field_or_exp, NestedField)\
                or isinstance(field_or_exp, ReverseNestedField):
                current_agg = current_agg.bucket(**field_or_exp.nested_params())

            elif field_or_exp.choices:
                params = field_or_exp.bucket_params()
                params['agg_type'] = 'terms'
                if self.default_size:
                    params['default_size'] = self.default_size
                current_agg = current_agg.bucket(**params)

            elif isinstance(field_or_exp, Field):
                params = field_or_exp.bucket_params()

                # /!\ TODO: not working with two group_by
                if self._order_by:
                    # If we're at the latest group_by, we can order by
                    if idx == last_idx:
                        params['order'] = self._order_by
                    # If the order_by is a doc count, we can add it at any level
                    elif self._order_by.keys() == ['_count']:
                        params['order'] = self._order_by

                params['agg_type'] = 'terms'
                if self.default_size:
                    params['size'] = self.default_size
                current_agg = current_agg.bucket(**params)

            else:
                raise NotImplementedError

        return current_agg

    def _configure_values(self, agg):
        for key, expression in self._expressions.items():
            if expression.is_field_agg():
                op = expression.__class__.__name__.lower()
                es_metric_name = key
                agg.metric(
                    es_metric_name,
                    op,
                    field=expression.field.get_storage_field(),
                    **expression.params
                )

    def _flatten_result(self, result):
        lines = flatten_result(result)

        key_to_field = {}
        for key, exp in self._expressions.items():
            if exp.is_doc_count():
                continue
            key_to_field[key] = exp

        for field_or_exp in self._group_by:
            if isinstance(field_or_exp, Aggregate):
                key_to_field[field_or_exp.field.storage_field] = field_or_exp
            else:
                key_to_field[field_or_exp.storage_field] = field_or_exp

        pretty_lines = []
        for line in lines:
            pretty_line = line.copy()
            self._add_computed_results(pretty_line)

            for key, value in pretty_line.items():
                if key in key_to_field:
                    field = key_to_field[key]
                    pretty_line[key] = field.get_casted_value(value)

            pretty_lines.append(pretty_line)

        return pretty_lines

    def _add_computed_results(self, line):
        computed_expressions = []

        while True:
            expressions_to_compute = []

            for key, expression in self._expressions.items():
                if not expression.is_computed():
                    continue

                if key in computed_expressions:
                    continue

                expressions_to_compute.append((key, expression))

            if not expressions_to_compute:
                break

            for key, expression in expressions_to_compute:
                try:
                    line[key] = expression.compute_one(line)
                    computed_expressions.append(key)
                except KeyError:
                    pass

    def _add_missing_lines(self, result, lines):
        enums = self._get_field_enums(result, lines)

        keys = list(product(*enums)) if enums else []
        if len(keys) == len(lines):
            return lines

        group_by_keys_without_nested = self._group_by_keys(nested=False)
        # We cast everything as str for easier matching, it won't change the type of keys in lines
        treated_hashes = [
            u','.join([unicode(line[key]) for key in group_by_keys_without_nested])
            for line in lines
        ]
        treated_hashes = set(treated_hashes)
        missing_keys = [
            key
            for key in keys
            if u','.join([unicode(k) for k in key]) not in treated_hashes
        ]

        lines += self._create_missing_lines(
            missing_keys,
            group_by_keys_without_nested,
        )

        return lines

    def _get_field_enums(self, result, lines):
        enums = []

        for field in self._group_by:
            if isinstance(field, Aggregate):
                if field.choice_keys():
                    enums.append(field.choice_keys())
                elif field.field.choices:
                    enums.append(field.field.choice_keys())
                else:
                    # We just add the lines' values
                    key = field.field.key
                    values = set([line[key] for line in lines])
                    values = sorted(list(values))
                    enums.append(values)

            elif field.is_range():
                enums.append(field.choice_keys())

            else:
                if field.choices:  # range / keyed?
                    enums.append(field.choice_keys())

                else:
                    # /!\ uncommon terms (beyond 10 first) will not be fetched by default
                    if 'custom_choices' in field.data:  #  and self.default_size == 0
                        # We add the custom choices
                        enums.append([c for c in field.data['custom_choices']])
                    else:
                        # We add the lines' values
                        key = field.key
                        values = set([line[key] for line in lines])
                        values = sorted(list(values))
                        enums.append(values)

        return enums

    def _group_by_keys(self, nested=True):
        return calc_group_by_keys(self._group_by, nested)

    def _create_missing_lines(self, missing_keys, group_by_keys):
        lines = []

        for missing_key in missing_keys:
            base_line = {}
            for idx, group_by_key in enumerate(group_by_keys):
                current_key = group_by_keys[idx]
                value = missing_key[idx]
                if hasattr(value, 'original_value'):  # In case of Choices
                    value = value.original_value
                base_line[current_key] = value

            lines.append(self._create_empty_line(base_line))

        return lines

    def _create_empty_line(self, base_line):
        empty_line = base_line.copy()

        for key, expression in self._expressions.items():
            empty_line[key] = None

        empty_line['doc_count'] = 0

        return empty_line



# -*- coding: utf-8 -*-

from six import with_metaclass

from elasticsearch_dsl import Mapping, Nested

from fiqs.exceptions import FieldError
from fiqs.fields import Field, NestedField


class ModelMetaClass(type):
    def __new__(cls, name, bases, attrs):
        klass = super(ModelMetaClass, cls).__new__(cls, name, bases, attrs)

        # We initialize the current class fields
        fields = []
        field_keys = []
        for field_key, field in attrs.items():
            if isinstance(field, Field):
                # We set the field's key
                field._set_key(field_key)
                # The field should be able to access the model
                field.model = klass

                fields.append(field)
                field_keys.append(field_key)

        # We do the same for the fields from the parent
        parents = [b for b in bases if isinstance(b, ModelMetaClass)]
        for p in parents:
            for field in p._fields:
                if field.key in field_keys:
                    msg = 'Field {} from class {} clashes with field ' +\
                          'with the same name in base class {}'
                    msg = msg.format(field.key, klass, p)
                    raise FieldError(msg)

                copy_field = field.get_copy()
                copy_field.model = klass

                setattr(klass, field.key, copy_field)

                fields.append(copy_field)
                field_keys.append(copy_field.key)

        klass._fields = fields

        return klass


class Model(with_metaclass(ModelMetaClass, object)):
    index = None
    doc_type = None

    @classmethod
    def get_index(cls, *args, **kwargs):
        if not cls.index:
            raise NotImplementedError('Model class should define an index')

        return cls.index

    @classmethod
    def get_doc_type(cls, *args, **kwargs):
        if not cls.doc_type:
            raise NotImplementedError('Model class should define a doc_type')

        return cls.doc_type

    @classmethod
    def get_mapping(cls):
        m = Mapping(cls.get_doc_type())
        m.meta('dynamic', 'strict')

        nested_mappings = {}
        fields_to_nest = []

        for field in cls._fields:
            if isinstance(field, NestedField):
                nested_mappings[field.key] = (field, Nested())

            if field.parent:
                fields_to_nest.append(field)
            else:
                m.field(field.storage_field, field.type)

        for field in fields_to_nest:
            if field.parent not in nested_mappings:
                raise Exception(
                    'Nested field {} needs to be defined in {}'.format(
                        field.parent, str(cls),
                    )
                )

            _, nested_mapping = nested_mappings[field.parent]
            nested_mapping.field(field.storage_field, field.type)

        for field, nested_mapping in nested_mappings.values():
            if not field.parent:
                m.field(field.key, nested_mapping)
            else:
                _, parent_mapping = nested_mappings[field.parent]
                parent_mapping.field(field.key, nested_mapping)

        return m
